package main

import (
	"bytes"
	"context"
	"fmt"
	"log"
	"os"
	"strconv"
	"strings"
	"time"

	"io/ioutil"
	"math/rand"
	"net/http"

	gabs "github.com/Jeffail/gabs/v2"
	"golang.org/x/net/proxy"
	irc "gopkg.in/irc.v3"
	zmq "gopkg.in/zeromq/goczmq.v4"
)

var default_nicks = []string{"Leopard", "SailfinMolly", "NarwhalMan", "BluejayMan", "GiantCaveSwallow", "GreatHornedOwl", "SnappingTurtleMan", "Wolverine", "GiantAardvark", "SeaLamprey", "GrasshopperMan", "WhitebrowedGibbon", "GiantLouse", "RedwingedBlackbird", "Fairy", "LorikeetMan", "GiantSkunk", "GiantAlligator", "CaveFish", "Mosquito", "GiantFox", "SaltwaterCrocodileMan", "GiantOsprey", "Lion", "WeaselMan", "Human", "Swordfish", "ClownLoach", "DragonflyMan", "GiantMosquito", "SpongeMan", "GiantCopperheadSnake", "GiantBeetle", "KestrelMan", "BullShark", "AardvarkMan", "PeachfacedLovebirdMan", "GiantThrips", "Capybara", "Rutherer", "MaskedLovebirdMan", "KnuckleWorm", "OnehumpedCamelMan", "Deer", "Cassowary", "BaskingShark", "Wombat", "GiantSkink", "GiantMongoose", "EchidnaMan", "ParakeetMan", "HareMan", "BlackcrestedGibbon", "VultureMan", "Slug", "BushmasterMan", "GiantBobcat", "GiganticPanda", "CapybaraMan", "MagmaMan", "SquidMan", "GiantCrab", "ChinchillaMan", "GiantOstrich", "GiantWhiteStork", "GiantDeer", "RedPanda", "Siamang", "Dwarf", "GiantLeopard", "Bobcat", "Duck", "TwohumpedCamelMan", "GiantBeaver", "Troglodyte", "GiantRattlesnake", "RoachMan", "Satyr", "Albatross", "CheetahMan", "SpottedRatfish", "AdderMan", "GrayLangur", "CockatielMan", "GiantMuskox", "GiantKea", "CaveBlob", "GiantWren", "SeaNettleJellyfish", "GiantMoth", "GiantDingo", "SpiderMonkeyMan", "Oyster", "GiantMaskedLovebird", "GiantSaltwaterCrocodile", "GiantKoala", "BrownRecluseSpiderMan", "MoonSnailMan", "GiantMagpie", "FoxSquirrel", "GiantLionTamarin", "Halibut", "Clownfish", "BobcatMan", "Groundhog", "DarkGnome", "GiantCapuchin", "GiantPlatypus", "HamsterMan", "BrownRecluseSpider", "Harpy", "BloodGnat", "Molemarian", "GroundhogMan", "GiantGrayLangur", "Beetle", "GiantChinchilla", "Buzzard", "BlackMamba", "BeakDog", "Magpie", "DesertTortoiseMan", "ChameleonMan", "IguanaMan", "TigerMan", "BarkScorpion", "GiantAlbatross", "GiantTiger", "Guppy", "GiantSnappingTurtle", "Mantis", "Raccoon", "Cardinal", "RhesusMacaque", "RatMan", "GiantOcelot", "GiantGrizzlyBear", "MountainGoatMan", "Carp", "Beaver", "Tiger", "GiantOtter", "ToadMan", "GreatBarracuda", "MonarchButterfly", "Thrips", "SlugMan", "GiantCaveToad", "GreenTreeFrog", "Moghopper", "GiantKingsnake", "KingCobraMan", "SpermWhaleMan", "Capuchin", "PondGrabber", "MooseCow", "Vulture", "RhinocerosMan", "BlackMambaMan", "GiantBat", "GiantAxolotl", "WhitetipReefShark", "GiantOrca", "Louse", "GrayLangurMan", "Snail", "Tapir", "Herring", "Nautilus", "SeaOtter", "Anole", "SkunkMan", "TwohumpedCamel", "GiantMooseBull", "RedPandaMan", "Damselfly", "Orca", "FlyingSquirrel", "KoalaMan", "GreenDevourer", "WildBoar", "Ibex", "GiantLion", "SeaSerpent", "ReptileMan", "LizardMan", "GiantNarwhal", "GiantMandrill", "Perch", "Oriole", "Hamster", "GiantRat", "GiantRaven", "Wagon", "GiantLizard", "GiantCougar", "Axolotl", "GiantWeasel", "WrenMan", "Creature", "GiantCardinal", "Kobold", "GiantEarthworm", "Cougar", "SnowyOwl", "BarnOwlMan", "PuffinMan", "GiantCoyote", "Hake", "GiantNautilus", "HornbillMan", "Bilou", "LeopardSealMan", "GreyParrotMan", "Unicorn", "GiantOpossum", "Stingray", "Bat", "Kangaroo", "GiantMooseCow", "PurringMaggot", "Impala", "CaveFloater", "Horse", "SwanMan", "Gazelle", "FoulBlendec", "LouseMan", "Elk", "Llama", "PorcupineMan", "PeregrineFalcon", "Lorikeet", "NakedMoleDog", "CaveSpider", "MagpieMan", "CrabMan", "GiantGilaMonster", "GiantCapybara", "Lizard", "HoaryMarmotMan", "WolfMan", "Kingsnake", "CapuchinMan", "GiantAnaconda", "GiantHornbill", "GiantToad", "NautilusMan", "GiantGraySquirrel", "TapirMan", "GiantSnowyOwl", "Bushtit", "GiantWalrus", "GiantOriole", "GiantParakeet", "HoneyBadger", "Cyclops", "SpinyDogfish", "Coati", "GiantRedSquirrel", "WhitespottedPuffer", "OspreyMan", "MothMan", "Chimpanzee", "Mink", "Walrus", "Tigerfish", "GiantDesertTortoise", "DeerMan", "BlueJay", "Sole", "Cat", "ElkMan", "Cow", "IceWolf", "Fly", "Crow", "BatRay", "MagmaCrab", "Angelshark", "Anchovy", "MonarchButterflyMan", "Bonobo", "BarnOwl", "GiganticTortoise", "Seahorse", "Badger", "GilaMonsterMan", "Hydra", "BarkScorpionMan", "Mussel", "WhaleShark", "FireSnake", "GrizzlyBearMan", "LongfinMakoShark", "BuzzardMan", "GrackleMan", "SaltwaterCrocodile", "ElephantSealMan", "Leech", "Koala", "OrcaMan", "LeopardGeckoMan", "GiantBarnOwl", "CongerEel", "GiantSwan", "ElephantSeal", "BlackhandedGibbon", "BeetleMan", "Alpaca", "MantisMan", "GiantSlothBear", "BloodMan", "RedSquirrelMan", "GiantEmu", "GiantCaveSpider", "OpossumMan", "PandaMan", "GiantMole", "Yeti", "OcelotMan", "ArmadilloMan", "Parakeet", "BlackBullhead", "RedSquirrel", "Ant", "Antman", "Merperson", "GiantBrownRecluseSpider", "GiantMantis", "GiantRoach", "GiantPorcupine", "GiantTortoiseWoman", "HyenaMan", "GiantRhesusMacaque", "Shad", "GiantBluejay", "Cavy", "GiantElk", "Hagfish", "Hedgehog", "CaveSwallow", "GiantMonarchButterfly", "MonitorLizard", "GiantWarthog", "HarpSealMan", "SlothBearMan", "Char", "Goose", "Crundle", "Gorlak", "Ocelot", "PondTurtle", "Hippo", "GiantHippo", "Aardvark", "Donkey", "Giraffe", "ThornbackRay", "GiantDamselfly", "Echidna", "LargeRat", "Elf", "Kea", "PileatedGibbon", "MinkMan", "Reindeer", "GiantAdder", "JumpingSpider", "LionMan", "Crab", "JumpingSpiderMan", "AnacondaMan", "GiantHyena", "GiantGreatHornedOwl", "AmethystMan", "SlothMan", "WolverineMan", "GiantRedwingedBlackbird", "Ayeaye", "Ostrich", "GiantKangaroo", "Yak", "Bluefish", "PythonMan", "MoonSnail", "GiantKestrel", "CaveFishMan", "Loon", "Draltha", "BeaverMan", "GiantEagle", "KakapoMan", "Rat", "OrioleMan", "MosquitoMan", "GiantBushtit", "GiantChameleon", "GrizzlyBear", "TwoleggedRhinoLizard", "PeachfacedLovebird", "FlyingSquirrelMan", "CassowaryMan", "GiantGroundhog", "ForgottenBeast", "MooseBull", "CaveDragon", "SilveryGibbon", "GiantKiwi", "HoaryMarmot", "Guineafowl", "LargeRoach", "GiantSloth", "SlothBear", "GiantLoon", "CapHopper", "GiantKingCobra", "Puffin", "GiantAnole", "Octopus", "LionTamarin", "PenguinMan", "FireMan", "CuttlefishMan", "CreepingEye", "Bumblebee", "Coyote", "Wren", "GiantPolarBear", "KeaMan", "Jackal", "Adder", "MonitorLizardMan", "Milkfish", "HorseshoeCrabMan", "HarpSeal", "BrownBullhead", "MountainGoat", "FrillShark", "GiantSponge", "Termite", "MountainGnome", "ChipmunkMan", "GiantGrasshopper", "Sasquatch", "Dragon", "Bugbat", "WhiteStork", "OctopusMan", "EmuMan", "GiantGrackle", "PondTurtleMan", "CoyoteMan", "HammerheadShark", "Swan", "GiantPangolin", "KingCobra", "LynxMan", "Weasel", "GilaMonster", "Skunk", "AlligatorMan", "DemonRat", "Panda", "Chinchilla", "RainbowTrout", "GiantTick", "FoxMan", "GiantPenguin", "PhantomSpider", "Opah", "Iguana", "MuskoxMan", "Hornbill", "TooSmallForPressurePlates", "GiantLeopardGecko", "BronzeColossus", "Goat", "Mule", "RattlesnakeMan", "WarthogMan", "KiwiMan", "GiantVulture", "GiantLorikeet", "WalrusMan", "RedwingedBlackbirdMan", "Eagle", "Olm", "DamselflyMan", "MooseMan", "DesertTortoise", "Stoat", "CardinalMan", "BushtitMan", "Python", "GiantPondTurtle", "Ogre", "GiantPython", "GiantIguana", "EmperorPenguin", "GiantLeopardSeal", "GrayGibbon", "WormMan", "BluePeafowl", "GiantDragonfly", "Kestrel", "GiantWolverine", "GiantOlm", "GiantPeregrineFalcon", "CopperheadSnake", "GreenTreeFrogMan", "GreyParrot", "FlyMan", "WaterBuffalo", "GiantEchidna", "GiantWolf", "BrookLamprey", "PolarBearMan", "GiantBuzzard", "Pig", "Troll", "SeaMonster", "GiantTwohumpedCamel", "CoatiMan", "IbexMan", "GreatHornedOwlMan", "GiantCheetah", "WhiteStorkMan", "Chipmunk", "MudMan", "GiantCassowary", "SparrowMan", "RodentMan", "PangolinMan", "StoatMan", "KangarooMan", "Sparrow", "WombatMan", "GiantBlackMamba", "CaveCrocodile", "GiantCuttlefish", "MaskedLovebird", "GiantAyeaye", "Skink", "Alligator", "GiantSpiderMonkey", "GiantCoati", "ElkBird", "MandrillMan", "Sloth", "OstrichMan", "Gorilla", "Mandrill", "LeechMan", "Tick", "GiantKakapo", "GiantHoaryMarmot", "VoraciousCaveCrawler", "Lungfish", "GiantTortoise", "OtterMan", "Minotaur", "GiantMountainGoat", "GiantElephantSeal", "TickMan", "ElephantSealWoman", "GiantHedgehog", "CreepyCrawler", "MooseWoman", "Sponge", "GiantCockatiel", "Orangutan", "BlackBear", "LeopardSeal", "GiantLynx", "IronMan", "Coelacanth", "SpermWhale", "JackalMan", "Porcupine", "SteelheadTrout", "BlindCaveBear", "PlumpHelmetMan", "NurseShark", "HungryHead", "SpottedWobbegong", "CommonSkate", "Dragonfly", "Mackerel", "RaccoonMan", "GiantGazelle", "RhesusMacaqueMan", "GiantFly", "AcornFly", "AlbatrossMan", "MantaRay", "OceanSunfish", "GiantFlyingSquirrel", "WildBoarMan", "Pixie", "Sheep", "Fox", "HoneyBee", "GiantGrouper", "GreatWhiteShark", "Narwhal", "SerpentMan", "BadgerMan", "Grimeling", "Chicken", "GiantHamster", "GiantBushmaster", "Jabberer", "MongooseMan", "Moth", "Strangler", "Drunian", "GabbroMan", "BluefinTuna", "GiantJaguar", "Hare", "Muskox", "Rhinoceros", "GiantTiercelPeregrine", "Grackle", "Giant", "CaveSwallowMan", "GazelleMan", "TooBigForPressurePlates", "GiantHarpSeal", "Mongoose", "TigerShark", "Squid", "Salmon", "Dog", "Dingo", "GiantOnehumpedCamel", "GiantLeech", "BlindCaveOgre", "CommonSnappingTurtle", "Jaguar", "KingsnakeMan", "GraySquirrel", "Goblin", "BlacktipReefShark", "GiantHare", "Penguin", "Armadillo", "Pangolin", "GiantChipmunk", "LongnoseGar", "RavenMan", "HelmetSnake", "GiantPeachfacedLovebird", "Flounder", "Turkey", "Manera", "GiantHorseshoeCrab", "AyeayeMan", "AlligatorSnappingTurtle", "Worm", "LittlePenguin", "CopperheadSnakeMan", "PlatypusMan", "Toad", "Lynx", "BlizzardMan", "GiantHoneyBadger", "LoonMan", "Glasseye", "HoneyBadgerMan", "GiantGreyParrot", "LionTamarinMan", "Grasshopper", "Cuttlefish", "HippoMan", "Chameleon", "Anaconda", "GiantRaccoon", "GiantBlackBear", "RiverOtter", "HedgehogMan", "GiantMoonSnail", "CougarMan", "Nightwing", "Osprey", "GiantSpermWhale", "OlmMan", "GiantTapir", "GiantRhinoceros", "ImpalaMan", "Warthog", "SkinkMan", "Roc", "Emu", "GiantRedPanda", "FireflyMan", "GiantWildBoar", "BlueShark", "Cod", "Wolf", "GiantJackal", "HorseshoeCrab", "Raven", "GiantSparrow", "GiantWombat", "YellowBullhead", "Platypus", "AmphibianMan", "LeopardMan", "GiantSlug", "GiantArmadillo", "Reacher", "Ettin", "JaguarMan", "FluffyWambler", "TooBigForBridges", "EagleMan", "CaveLobster", "GiantMonitorLizard", "Kiwi", "CrowMan", "GiantStoat", "Kakapo", "GiraffeMan", "Rattlesnake", "OnehumpedCamel", "GiantJumpingSpider", "GiganticSquid", "Elephant", "SnowyOwlMan", "ThripsMan", "Cheetah", "GiantElephant", "SnailMan", "ShortfinMakoShark", "GiantBarkScorpion", "Gremlin", "BatMan", "Opossum", "FireImp", "FleshBall", "Pike", "BlackBearMan", "GiantPuffin", "PolarBear", "ElephantMan", "SpiderMonkey", "FloatingGuts", "PeregrineFalconMan", "GiantImpala", "AxolotlMan", "GiantFirefly", "Hyena", "Sturgeon", "WhitehandedGibbon", "GiantSnail", "GiantOctopus", "Marlin", "GiantGreenTreeFrog", "Bushmaster", "Cockatiel", "GiantBadger", "GiantCrow", "Rabbit", "GiantMink", "BandedKnifefish", "LeopardGecko", "AnoleMan", "GraySquirrelMan", "Firefly", "GiantIbex", "GiantGiraffe", "GiantTortoiseMan", "DingoMan"}
var RNG = rand.New(rand.NewSource(time.Now().UnixNano()))

type BotConfig struct {
	Nick         string
	Server       string
	Channel      string
	Victim       string
	OpMode       bool
	GreetChance  int
	SuffixLength int
	Friends      []string
	HealthPort   int
	LogIrc       bool // log irc traffic to stdout
}

type BotStatus struct {
	Connected bool
	Invites   int
}

type BotState struct {
	IRCClient  *irc.Client
	Subscriber *zmq.Sock
}

type Bot struct {
	Config *BotConfig
	Status *BotStatus
	State  *BotState
}

func NewBotConfigFromEnv() *BotConfig {
	c := &BotConfig{}

	c.Server = os.Getenv("IRC_SERVER")
	if c.Server == "" {
		log.Fatal("Set IRC_SERVER!")
	}

	c.Channel = os.Getenv("IRC_CHANNEL")
	if c.Channel == "" {
		log.Fatal("Set IRC_CHANNEL!")
	}

	enable_op_mode := os.Getenv("IRC_OP_MODE")
	if enable_op_mode != "" {
		c.OpMode = true
		log.Println("op mode enabled")
	}

	c.SuffixLength = getNumericVar("SUFFIX_DIGITS", 0)
	if c.SuffixLength < 0 || c.SuffixLength > 10 {
		panic(fmt.Errorf("invalid SUFFIX_DIGITS: %d", c.SuffixLength))
	}

	c.GreetChance = getNumericVar("GREET_CHANCE", 0)
	if c.GreetChance > 1000 || c.GreetChance < 0 {
		log.Fatalf("invalid GREET_CHANCE: '%d'", c.GreetChance)
	}

	c.HealthPort = getNumericVar("HEALTH_PORT", 8000)
	if c.HealthPort < 0 || c.HealthPort > 65535 {
		panic(fmt.Errorf("invalid HEALTH_PORT: %d", c.HealthPort))
	}

	c.Victim = os.Getenv("IRC_VICTIM")
	c.LogIrc = os.Getenv("LOG_IRC") != ""

	c.Nick = genNick(c.SuffixLength, c.OpMode)

	botconfig_loadFriends(c)
	return c
}

func botconfig_loadFriends(c *BotConfig) {
	c.Friends = make([]string, 0)

	friendsV := os.Getenv("IRC_EXTRAFRIENDS")
	if friendsV == "" {
		return
	}
	for _, name := range strings.Split(strings.TrimSpace(friendsV), ",") {
		name := strings.TrimSpace(name)
		if name != "" {
			log.Printf("loaded extra friend: %s", name)
			c.Friends = append(c.Friends, name)
		}
	}
}

func loadNames() {
	namesPath := os.Getenv("NAMES_FPATH")
	if namesPath == "" {
		return
	}

	content, err := ioutil.ReadFile(namesPath)
	if err != nil {
		panic(err)
	}
	default_nicks = []string{}
	for _, line := range strings.Split(string(content), "\n") {
		stripped := strings.TrimSpace(line)
		if stripped != "" {
			default_nicks = append(default_nicks, stripped)
		}
	}
	if len(default_nicks) == 0 {
		panic(fmt.Errorf("no names loaded from %s", namesPath))
	}
}

func getNumericVar(key string, default_ int) int {
	varS := os.Getenv(key)
	if varS != "" {
		var err error
		default_, err = strconv.Atoi(varS)
		if err != nil {
			panic(fmt.Errorf("invalid %s: '%s'", key, varS))
		}
	}
	return default_
}

func friendlyNick(nick string, pool []string) bool {
	for _, n := range pool {
		if nick == n {
			return true
		}
	}
	for _, n := range pool {
		if strings.HasPrefix(nick, n) {
			return true
		}
	}
	if strings.HasPrefix(nick, "Op") {
		return friendlyNick(nick[2:], pool)
	}
	return false
}

func (b *Bot) FriendlyNick(nick string) bool {
	return friendlyNick(nick, b.Config.Friends) || friendlyNick(nick, default_nicks)
}

func genNick(digits int, opmode bool) string {
	n := default_nicks[RNG.Intn(len(default_nicks))] + trailingDigits(digits)
	if opmode {
		n = "Op" + n
	}
	return n
}

func trailingDigits(length int) string {
	trailing := ""
	for ; length > 0; length-- {
		trailing = fmt.Sprintf("%s%d", trailing, RNG.Intn(10))
	}
	return trailing
}

func NewBotFromEnv() *Bot {
	b := &Bot{
		Config: NewBotConfigFromEnv(),
		Status: &BotStatus{
			Connected: false,
			Invites:   0,
		},
		State: &BotState{},
	}
	return b
}

func (b *Bot) Run(ctx context.Context) (string, error) {
	go b.runBackground()
	go b.runStatusHandler()

	log.Println("dialing...")

	dialCtx, cancelDial := context.WithTimeout(context.Background(), time.Second*10)
	defer cancelDial()

	conn, err := proxy.Dial(dialCtx, "tcp", b.Config.Server)
	if err != nil {
		log.Fatal(err)
	}
	hostname, _ := os.Hostname()
	config := irc.ClientConfig{
		Nick:    b.Config.Nick,
		User:    hostname,
		Name:    "dabbing on " + b.Config.Victim,
		Handler: irc.HandlerFunc(b.handleIRC),
	}
	log.Println("starting client...")
	b.State.IRCClient = irc.NewClient(conn, config)
	err = b.State.IRCClient.Run()
	return "i died", err
}

func (b *Bot) runBackground() {
	publisher_address := os.Getenv("PUBLISHER_ADDRESS")
	if publisher_address == "" {
		return
	}

	sub, err := zmq.NewSub(fmt.Sprintf("tcp://%s", publisher_address), "_all")
	if err != nil {
		log.Fatal(err)
	}
	defer sub.Destroy()

	channel := strings.ToLower(b.Config.Nick)
	log.Printf("channel: %s", channel)
	sub.SetSubscribe(channel)

	if b.Config.OpMode {
		sub.SetSubscribe("_ops")
	} else {
		sub.SetSubscribe("_nops")
	}

	b.State.Subscriber = sub
	log.Println("subscriber bound")

	for {
		resp, err := sub.RecvMessage()
		if err != nil {
			log.Printf("Sub recv failed: %s", err)
			continue
		}
		msg := bytes.Join(resp, nil)

		parts := bytes.SplitN(msg, []byte{' '}, 2)

		// channel := string(parts[0])
		message := parts[1]

		// log.Printf("channel: '%s'", channel)
		// log.Printf("message: '%s'", message)
		// log.Printf("len    : %d", len(message))

		b.handleBackgroundMessage(message)

		// log.Printf("")
	}
}

func roll(chance int) bool {
	if chance <= 0 {
		return false
	}
	return RNG.Intn(chance) == 0
}

func (b *Bot) handleBackgroundMessage(message []byte) {
	json, err := gabs.ParseJSON(message)
	if err != nil {
		log.Printf("invalid message: %s", message)
		return
	}

	chance, ok := json.S("chance").Data().(float64)
	if ok && !roll(int(chance)) {
		log.Printf("chanced out")
		return
	}

	// todo replace this with smarter json struct nonsense
	for key, child := range json.ChildrenMap() {
		switch key {
		case "greet_chance":
			loadIntField(&b.Config.GreetChance, child)
			log.Printf("setting GreetChance to: %d", b.Config.GreetChance)
		case "victim":
			loadStrField(&b.Config.Victim, child)
			log.Printf("setting Victim to: %s", b.Config.Victim)
		case "join":
			if !b.Status.Connected {
				continue
			}
			channel, ok := child.Data().(string)
			if !ok || channel == "" {
				continue
			}
			b.State.IRCClient.WriteMessage(&irc.Message{
				Command: "JOIN",
				Params: []string{
					channel,
				},
			})
		case "part":
			if !b.Status.Connected {
				continue
			}
			channel, ok := child.Data().(string)
			if !ok || channel == "" {
				continue
			}
			b.State.IRCClient.WriteMessage(&irc.Message{
				Command: "PART",
				Params: []string{
					channel,
					"",
				},
			})
		case "say":
			if !b.Status.Connected {
				continue
			}
			sayMsg, ok := child.Data().(string)
			if !ok || sayMsg == "" {
				continue
			}
			sayTo, ok := json.S("to").Data().(string)
			if !ok || sayTo == "" {
				sayTo = b.Config.Channel
			}
			b.State.IRCClient.WriteMessage(&irc.Message{
				Command: "PRIVMSG",
				Params: []string{
					sayTo, sayMsg,
				},
			})
		case "op":
			if !b.Status.Connected {
				continue
			}
			opNick, ok := child.Data().(string)
			if !ok || opNick == "" {
				continue
			}
			b.State.IRCClient.WriteMessage(&irc.Message{
				Command: "MODE",
				Params: []string{
					b.Config.Channel,
					"+o",
					opNick,
				},
			})
		case "nick":
			if !b.Status.Connected {
				continue
			}
			b.DoNewNick()
		case "name":
			if !b.Status.Connected {
				continue
			}
			b.DoNewNick()
		case "die":
			panic("by request")
		case "chance":
		case "to":
		default:
			log.Printf("unknown var field %s", key)
		}
	}
}

func loadIntField(dest *int, ob *gabs.Container) {
	value, ok := ob.Data().(float64)
	if !ok {
		log.Printf("not int: %v", ob)
		return
	}
	*dest = int(value)
}

func loadStrField(dest *string, ob *gabs.Container) {
	value, ok := ob.Data().(string)
	if !ok {
		log.Printf("not str: %v", ob)
		return
	}
	*dest = value
}

func (b *Bot) DoNewNick() {
	nick := genNick(b.Config.SuffixLength, b.Config.OpMode)
	b.UpdateNick(nick)
	b.State.IRCClient.Write("NICK " + nick)
}

func (b *Bot) runStatusHandler() {

	healthAddress := fmt.Sprintf("0.0.0.0:%d", b.Config.HealthPort)
	log.Printf("serving health on %s", healthAddress)
	http.HandleFunc("/health", b.statusHandler())
	log.Fatal(http.ListenAndServe(healthAddress, nil))
}

func (b *Bot) statusHandler() func(http.ResponseWriter, *http.Request) {
	return func(w http.ResponseWriter, r *http.Request) {
		if !b.Status.Connected {
			http.Error(w, "Not connected", http.StatusInternalServerError)
			return
		}
		w.Write([]byte("ok"))
	}
}

func (b *Bot) UpdateNick(nick string) {
	b.State.Subscriber.SetUnsubscribe(strings.ToLower(b.Config.Nick))
	b.State.Subscriber.SetSubscribe(strings.ToLower(nick))
	b.Config.Nick = nick
}

func (b *Bot) handleIRC(c *irc.Client, m *irc.Message) {
	if b.Config.LogIrc {
		log.Printf("%+v\n", m)
	}

	if m.Command == "001" { // Handle connection established
		b.Status.Connected = true
		// 001 is welcome event
		c.Write("JOIN " + b.Config.Channel)

	} else if m.Command == "PRIVMSG" && c.FromChannel(m) { // Handle chat message to print stats

		text := strings.Split(m.Trailing(), " ")
		if text[0] == c.CurrentNick()+":" && text[len(text)-1] == "stats" {
			c.WriteMessage(&irc.Message{
				Command: "PRIVMSG",
				Params: []string{
					m.Params[0],
					"I've sent " + strconv.FormatInt(int64(b.Status.Invites), 10) + " invites. Current target: " + b.Config.Victim + " :)",
				},
			})
		} else if text[len(text)-1] == "name" {
			b.DoNewNick()
		} else if text[len(text)-1] == "invite" && b.Config.Victim != "" {
			c.Write("INVITE " + b.Config.Victim + " " + b.Config.Channel)
			b.Status.Invites++
		} /* else if strings.Contains(m.Trailing(), "greetings") {
			c.WriteMessage(&irc.Message{
				Command: "PRIVMSG",
				Params: []string{
					m.Params[0],
					config_victim + ": hi",
				},
			})
		}*/
	} else if m.Command == "JOIN" && c.FromChannel(m) {
		if m.Prefix.Name != c.CurrentNick() {
			if b.Config.GreetChance == 1000 || RNG.Intn(1000) < b.Config.GreetChance {
				c.WriteMessage(&irc.Message{
					Command: "PRIVMSG",
					Params: []string{
						m.Params[0],
						"greetings " + m.Prefix.Name,
					},
				})
			}
			if b.Config.OpMode && b.FriendlyNick(m.Prefix.Name) {
				log.Printf("opping %s", m.Prefix.Name)
				c.WriteMessage(&irc.Message{
					Command: "MODE",
					Params: []string{
						m.Params[0],
						"+o",
						m.Prefix.Name,
					},
				})
			}
		}
	}
}

func main() {
	loadNames()
	b := NewBotFromEnv()
	log.Printf("nick: %s", b.Config.Nick)
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()
	ret, err2 := b.Run(ctx)
	if err2 != nil {
		log.Fatalln(err2)
	}
	log.Println(ret)
}
